﻿// © Benn Eichhorn 2010
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Algobox.Feeds.Finance.Chix.CHIXMD.Messages
{
    /// <summary>
    /// Singleton Chi-X Order Add Message
    /// </summary>
    public class OrderAdd : IMessage
    {
        private OrderAdd() { }

        private static OrderAdd m_Instance = new OrderAdd();

        /// <summary>
        /// Get's the singleton instance of the OrderAdd class
        /// </summary>
        /// <returns>Current OrderAdd</returns>
        public static OrderAdd GetInstance() { return m_Instance; }

        
        public OrderAdd(DateTime timeStamp, int orderReference, bool isBuy, ulong shares, string stock, double price)
        {
            OrderReference = orderReference;
            IsBuy = isBuy;
            Shares = shares;
            Stock = stock;
            Price = price;
        }
        

        public char MessageType { get { return 'S'; } }

        /// <summary>
        /// Milliseconds since midnight
        /// </summary>
        public int TimeStampMillis;

        /// <summary>
        /// Slow, calculates current TimeStamp from TimeStampMillis
        /// </summary>
        public DateTime TimeStamp { get { return MessageParser.Date.AddMilliseconds(TimeStampMillis); } }

        /// <summary>
        /// Day unique order reference number
        /// </summary>
        public int OrderReference;

        /// <summary>
        /// True for buy orders, false for sell orders
        /// </summary>
        public bool IsBuy;
                
        /// <summary>
        /// Total number of shares being added to the book (may be less than the number of shares entered because part of the order may trade before being posted to the book)
        /// </summary>
        public ulong Shares;

        /// <summary>
        /// Stock symbol right padded with spaces
        /// </summary>
        public string Stock;

        /// <summary>
        /// price as it comes from exchange, WARNING this can be in long (7) or short (4) format
        /// </summary>
        private long PriceRaw;

        /// <summary>
        /// The display price of the order
        /// </summary>
        public double Price;


        public override string ToString()
        {
            return String.Format("IMessage->OrderAdd: Stock={0},IsBuy={1},Shares={2},Price={3},OrderReference={4},TimeStamp={5}", Stock, IsBuy, Shares, Price, OrderReference, TimeStamp);
        }

        // Order Add
        const int S_A_LEN = 43;
        const int S_A_ORDERREF_LOC = 10;
        const int S_A_ORDERREF_LEN = 9;
        const int S_A_BUYSELL_LOC = 19;
        const int S_A_SHARES_LOC = 20;
        const int S_A_SHARES_LEN = 6;
        const int S_A_STOCK_LOC = 26;
        const int S_A_STOCK_LEN = 6;
        const int S_A_PRICE_LOC = 32;
        const int S_A_PRICE_LEN = 10;
        const int S_A_PRICE_DECIMAL_DIVISOR = 10000;

        // Order Add Long
        const int S_a_LEN = 56;
        const int S_a_ORDERREF_LOC = 10;
        const int S_a_ORDERREF_LEN = 9;
        const int S_a_BUYSELL_LOC = 19;
        const int S_a_SHARES_LOC = 20;
        const int S_a_SHARES_LEN = 10;
        const int S_a_STOCK_LOC = 30;
        const int S_a_STOCK_LEN = 6;
        const int S_a_PRICE_LOC = 36;
        const int S_a_PRICE_LEN = 19;
        const int S_a_PRICE_DECIMAL_DIVISOR = 10000000;
        

        internal static IMessage ParseShort(ref string inbound)
        {
            // get sequenced data message timestamp
            if (!Int32.TryParse(inbound.Substring(MessageParser.S_TIMESTAMP_LOC, MessageParser.S_TIMESTAMP_LEN), out m_Instance.TimeStampMillis))
            {
                Messages.Error message = Messages.Error.GetInstance("Sequenced data message failed to parse TimeStamp millis as Double", inbound);
                return message;
            }

            if (!UInt64.TryParse(inbound.Substring(S_A_SHARES_LOC, S_A_SHARES_LEN), out m_Instance.Shares))
            {
                Messages.Error message = Messages.Error.GetInstance("OrderAdd failed to parse Shares as Int32", inbound);
                return message;
            }
            if (!Int32.TryParse(inbound.Substring(S_A_ORDERREF_LOC, S_A_ORDERREF_LEN), out m_Instance.OrderReference))
            {
                Messages.Error message = Messages.Error.GetInstance("OrderAdd failed to parse OrderReference as Int32", inbound);
                return message;
            }
            if (!Int64.TryParse(inbound.Substring(S_A_PRICE_LOC, S_A_PRICE_LEN), out m_Instance.PriceRaw))
            {
                Messages.Error message = Messages.Error.GetInstance("OrderAdd failed to parse Price as Int64", inbound);
                return message;
            }
            
            m_Instance.IsBuy = inbound[S_A_BUYSELL_LOC] == 'B' ? true : false;

            m_Instance.Stock = inbound.Substring(S_A_STOCK_LOC, S_A_STOCK_LEN).TrimEnd();
            m_Instance.Price = m_Instance.PriceRaw / S_A_PRICE_DECIMAL_DIVISOR;

            return m_Instance;
        }

        internal static IMessage ParseLong(ref string inbound)
        {
            // get sequenced data message timestamp
            if (!Int32.TryParse(inbound.Substring(MessageParser.S_TIMESTAMP_LOC, MessageParser.S_TIMESTAMP_LEN), out m_Instance.TimeStampMillis))
            {
                Messages.Error message = Messages.Error.GetInstance("Sequenced data message failed to parse TimeStamp millis as Double", inbound);
                return message;
            }

            if (!UInt64.TryParse(inbound.Substring(S_a_SHARES_LOC, S_a_SHARES_LEN), out m_Instance.Shares))
            {
                Messages.Error message = Messages.Error.GetInstance("OrderAdd failed to parse Shares as Int32", inbound);
                return message;
            }
            if (!Int32.TryParse(inbound.Substring(S_a_ORDERREF_LOC, S_a_ORDERREF_LEN), out m_Instance.OrderReference))
            {
                Messages.Error message = Messages.Error.GetInstance("OrderAdd failed to parse OrderReference as Int32", inbound);
                return message;
            }
            if (!Int64.TryParse(inbound.Substring(S_a_PRICE_LOC, S_a_PRICE_LEN), out m_Instance.PriceRaw))
            {
                Messages.Error message = Messages.Error.GetInstance("OrderAdd failed to parse Price as Int64", inbound);
                return message;
            }

            m_Instance.IsBuy = inbound[S_a_BUYSELL_LOC] == 'B' ? true : false;

            m_Instance.Stock = inbound.Substring(S_a_STOCK_LOC, S_a_STOCK_LEN).TrimEnd();
            m_Instance.Price = m_Instance.PriceRaw / S_a_PRICE_DECIMAL_DIVISOR;

            return m_Instance;
        }

        /// <summary>
        /// Deep clone
        /// </summary>
        /// <returns></returns>
        public OrderAdd Clone()
        {
            OrderAdd clone = new OrderAdd();
            clone.IsBuy = m_Instance.IsBuy;
            clone.OrderReference = m_Instance.OrderReference;
            clone.TimeStampMillis = m_Instance.TimeStampMillis;
            clone.Price = m_Instance.Price;
            clone.PriceRaw = m_Instance.PriceRaw;
            clone.Shares = m_Instance.Shares;
            clone.Stock = m_Instance.Stock;
            return clone;
        }
    }
}
