﻿//
// © Algobox Ltd - Benn Eichhorn
//
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ComponentModel;
using Algobox.Structure.Finance.Strategies.MarketFiller;
using Algobox.Objects.Finance.Assets.Stock;
using System.Threading;

namespace Algobox.Structure.Finance.Strategies.FractionLive
{
    public class FractionUnit : INotifyPropertyChanged, IMarketFillerCallback, IStrategyUnit
    {
        #region Constructors

        public FractionUnit(
            FractionStock strategy,
            Algobox.Feeds.Finance.Trading.ITradingFeedAdaptor tradingFeed,
            string id,
            bool isBuy,
            double entryPriceTarget,
            double exitPrice,
            uint quantity,
            double _phase1rsi,
            double _phase2rsi,
            double _phase3rsi,
            double _phase4rsi,
            double _phase1price,
            double _phase3price)
        {
            _tradingFeed = tradingFeed;
            _strategy = strategy;
            _isBuy = isBuy;
            _priceEntryTarget = entryPriceTarget;
            _priceExitTarget = exitPrice;
            _quantityInitial = quantity;
            _quantityRemain = quantity;
            _id = id;
            _logId = _strategy.Security + "|" + id;
            _lock = new object();

            Phase1RSI = _phase1rsi;
            Phase2RSI = _phase2rsi;
            Phase3RSI = _phase3rsi;
            Phase4RSI = _phase4rsi;
            Phase1Price = _phase1price;
            Phase3Price = _phase3price;
                        
            if (_strategy.Security.Currency == Objects.Finance.Currencies.Currency.GBX)
                _divisor = 100;
            else
                _divisor = 1;

            if(Log.IsInfoEnabled)
                Log.Info("{0}|{1}|Creating FractionUnit of {2}@{3} with an initial target at {4}", strategy.Id, _logId, quantity, entryPriceTarget, exitPrice);

            Profit = CalculateProfit(exitPrice);
        }

        #endregion
        //////////
        #region Member Variables

        /// <summary>
        /// Logger used by this class
        /// </summary>        
        private static readonly NLog.Logger Log = NLog.LogManager.GetCurrentClassLogger();

        public event PropertyChangedEventHandler PropertyChanged;

        private object _lock;
        
        private Algobox.Feeds.Finance.Trading.ITradingFeedAdaptor _tradingFeed;
        private FractionStock _strategy;
        private string _id;
        private string _logId;
        private uint _quantityInitial;
        private uint _quantityRemain;
        private readonly bool _isBuy;
        private int _divisor;

        private double _priceEntryTarget;
        private double _priceEntryActual;
        private double _priceExitTarget;
        private double _priceExitActual;
        private double _profit;

        private MarketFillOrder _orderExit;
        
        DateTime _timeStampEntry;
        DateTime _timeStampExit;

        #endregion
        //////////
        #region Properties

        public string Id { get { return this.GetType().Name + "|" + _logId; } }
        public bool IsBuy { get { return _isBuy; } }
        public string Name { get { return _strategy.Name; } }
        public Strategy Strategy { get { return _strategy; } }
        public string TerrapinId { get { return _id; } }
        public Side Direction { get { return _isBuy ? Side.Long : Side.Short; } }
        public uint Quantity { get { return _quantityInitial; } private set { _quantityInitial = value; NotifyPropertyChanged("Quantity"); } }
        public uint Remaining { get { return _quantityRemain; } private set { _quantityRemain = value; NotifyPropertyChanged("Remaining"); } }
        public double EntryTarget { get { return _priceEntryTarget; } }
        public double EntryActual { get { return _priceEntryActual == 0 ? _priceEntryTarget : _priceEntryActual; } }
        
        public double ExitTarget 
        {
            get { return _priceExitTarget; }
            set
            {
                if (_priceExitTarget != value)
                {
                    if (Log.IsInfoEnabled)
                        Log.Info("{0}|{1}|ExitTarget Changing from {2} to {3}", _strategy.Id, _logId, _priceExitTarget, value);

                    _priceExitTarget = value;

                    NotifyPropertyChanged("ExitTarget");
                }
            } 
        }

        public double ExitActual { get { return _priceExitActual; } private set { _priceExitActual = value; NotifyPropertyChanged("ExitActual"); NotifyPropertyChanged("SlipExit"); } }

        public double Profit { get { return _profit; } private set { _profit = value; NotifyPropertyChanged("Profit"); } }

        public double SlipEntry { get { return _priceEntryActual == 0 ? 0 : _isBuy ? _priceEntryTarget - _priceEntryActual : _priceEntryActual - _priceEntryTarget; } }
        public double SlipExit { get { return _priceExitActual == 0 ? 0 : _isBuy ? _priceExitTarget - _priceExitActual : _priceExitActual - _priceExitTarget; } }
        public string TimeStampEntry { get { return _timeStampEntry.ToLongTimeString(); } }
        public string TimeStampExit { get { return _timeStampExit.ToLongTimeString(); } }

        public double Phase1RSI { get; private set; }
        public double Phase2RSI { get; private set; }
        public double Phase3RSI { get; private set; }
        public double Phase4RSI { get; private set; }
        public double Phase1Price { get; private set; }
        public double Phase3Price { get; private set; }

        #endregion
        //////////
        #region Functions
        
        public void RegisterEntry(double price, DateTime timstamp)
        {
            _priceEntryActual = price;
            _timeStampEntry = timstamp;
            NotifyPropertyChanged("EntryActual");
            NotifyPropertyChanged("SlipEntry");
            NotifyPropertyChanged("TimeStampEntry");
        }

        public void ExitUnit(DateTime timeStamp)
        {
            if (_orderExit == null)
            {
                if (_quantityRemain != 0)
                {
                    lock (_lock)
                    {
                        if (_orderExit == null && _quantityRemain != 0)
                        {
                            if (Log.IsInfoEnabled)
                                Log.Info("{0}|{1}|ExitUnit activating exit order at {2}", _strategy.Id, _logId, _priceExitTarget);

                            _orderExit = new MarketFillOrder(_tradingFeed, this, _strategy.Security, _strategy.Account, _isBuy, _priceExitTarget, _quantityInitial, Objects.Finance.Orders.Order.TimeInForce.IOC);
                            _orderExit.Start();
                            _timeStampExit = timeStamp;
                            NotifyPropertyChanged("TimeStampExit");
                        }
                    }
                }
                else
                {
                    Log.Error("{0}|ExitUnit Can't exit with 0 remaining quantity!", _logId);
#if DEBUG
                    throw new InvalidOperationException("Can't exit with 0 remaining quantity!");
#endif
                }
            }
            else
            {
                Log.Error("{0}|ExitUnit Attempting to exit unit more than once!", _logId);
#if DEBUG
                throw new InvalidOperationException("Should not exit more than once!");
#endif
            }
        }

        public void OnMarketFillerCompleted(MarketFillOrder caller)
        {
            if (caller == _orderExit)
            {
                lock (_lock)
                {
                    Profit = CalculateProfit(caller.AveragePrice);

                    if (Log.IsInfoEnabled)
                        Log.Info("{0}|Unit complete, exit order filled, achieved {1}@{2}", _logId, caller.Quantity, caller.AveragePrice);

                    ExitActual = caller.AveragePrice;
                    Remaining = 0;
                    _strategy.OnFractionUnitComplete(this);
                    _orderExit = null;
                    
                }
            }
            else
            {
                Log.Error("{0}|OnMarketFillerCompleted Unhandled MarketFillOrder({1})", _logId, caller.ToString());
            }
        }

        public void OnMarketFillerFailed(MarketFillOrder caller)
        {
            if (caller == _orderExit)
            {
                lock (_lock)
                {
                    if (caller.QuantityFilled > 0)
                    {
                        ExitActual = caller.AveragePrice;
                        Remaining -= caller.Quantity;

                        if (Log.IsFatalEnabled)
                            Log.Fatal("{0}|Terrapin exit leg partially failed we recieved only {1}@{2} leaving a difference of {3}", _logId, _quantityRemain, caller.AveragePrice);

                        _strategy.OnFractionUnitFailed(this);
                    }
                    else
                    {
                        if (Log.IsErrorEnabled)
                            Log.Error("{0}|Terrapin exit leg failed, disabling this Terrapin strategy leaving open position of {1}", _logId, _quantityRemain);

                        _strategy.OnFractionUnitFailed(this);
                    }                
                }         
            }
            else
            {
                Log.Error("{0}|OnMarketFillerCompleted Unhandled MarketFillOrder({1})", _logId, caller.ToString());
            }
        }

        public void OnBestPrice(double price, DateTime timeStamp)
        {
            if (Monitor.TryEnter(_lock, 10))
            {
                try
                {
                    if (_orderExit == null && _priceEntryActual != 0)
                    {
                        if (_isBuy && price <= _priceExitTarget)
                        {
                            if (Log.IsInfoEnabled)
                                Log.Info("{0}|Exiting unit as latest price {1} is <= exit price {2}", _logId, price, _priceExitTarget);

                            ExitUnit(timeStamp);
                        }
                        else if (!_isBuy && price >= _priceExitTarget)
                        {
                            if (Log.IsInfoEnabled)
                                Log.Info("{0}|Exiting unit as latest price {1} is >= exit price {2}", _logId, price, _priceExitTarget);

                            ExitUnit(timeStamp);
                        }
                    }

                    Profit = CalculateProfit(price);
                }
                finally
                {
                    Monitor.Exit(_lock);
                }
            }
        }

        private double CalculateProfit(double exitPrice)
        {
            double entryPrice;
            if (_priceEntryActual == 0)
                entryPrice = _priceEntryTarget;
            else
                entryPrice = _priceEntryActual;

            return _isBuy ?
                (exitPrice - entryPrice) * _quantityInitial / _divisor:
                (entryPrice - exitPrice) * _quantityInitial / _divisor;
        }

        private void NotifyPropertyChanged(string propertyName)
        {
            if (PropertyChanged != null)
            {
                PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
            }
        }
        
        #endregion
        //////////
        #region Enums


        #endregion


    }
}