﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Algobox.Toolbox.Collections;
using System.ComponentModel;
using Algobox.Objects.Finance.Assets.Stock;
using Algobox.Feeds.Finance.MarketData;

namespace Algobox.Structure.Finance.Strategies.FractionPrototype
{
    public class FractionComponentDay : Strategy, IFractionComponent, INotifyPropertyChanged, IMarketUpdateLast
    {
        public FractionComponentDay(
            Guid databaseId,
            FractionManagement manager,
            StockSecurity securityA,
            StockSecurity securityB)
            :base(securityA, securityB)
        {

            _databaseId = databaseId;
            Manager = manager;

            manager.MarketFeed.SubscribeLast(securityA, this);
            manager.MarketFeed.SubscribeLast(securityB, this);
            
            LOG.Info("{0}|{1}:{2}|FractionComponentDay with DatabaseId({3})", _strategyIdString, securityA, securityB, databaseId);
        }

        public event PropertyChangedEventHandler PropertyChanged;
        
        private static readonly NLog.Logger LOG = NLog.LogManager.GetCurrentClassLogger();


        private Guid _databaseId;
        private bool _isEnabled;
        private double _midA;
        private double _midB;
        private double _moveDayA;
        private double _moveDayB;
        private double _diffDay;
        private ushort _value = 0;
        private double _trigger = 0;
        private double _exit = 0;
        
        private FractionTrigger _leg;

        public FractionManagement Manager { get; set; }

        public bool IsTriggered { get { return _leg != null; } }

        public Guid DatabaseId { get { return _databaseId; } }
        public bool IsEnabled
        {
            get { return _isEnabled; }
            set
            {
                if (value)
                {
                    if (_trigger == 0 || _value == 0)
                    {
                        _isEnabled = false;
                    }
                    // or do other start up checks here
                    else
                    {
                        _isEnabled = true;
                    }
                }
                else
                {
                    // diable stuff here
                    _isEnabled = false;
                }
                NotifyPropertyChanged("IsEnabled");
            }
        }
        public string SymbolA
        {
            get { return _security.Symbol; }
            set
            {
                if (!String.IsNullOrWhiteSpace(value))
                {
                    Manager.MarketFeed.UnsubscribeLast(_security, this);
                    _security = Manager.Data.AddStock(value);
                    Manager.MarketFeed.SubscribeLast(_security, this);
                    NotifyPropertyChanged("SymbolA");
                }
            }
        }
        public string SymbolB
        {
            get { return _securityPair.Symbol; }
            set
            {
                if (!String.IsNullOrWhiteSpace(value))
                {
                    Manager.MarketFeed.UnsubscribeLast(_securityPair, this);
                    _securityPair = Manager.Data.AddStock(value);
                    Manager.MarketFeed.SubscribeLast(_securityPair, this);
                    NotifyPropertyChanged("SymbolB");
                }
            }
        }        
        public double CloseA
        {
            get { return _security.Prices.OHLC.Close; }
            set { _security.Prices.OHLC.Close = value;  NotifyPropertyChanged("CloseA"); }
        }
        public double CloseB
        {
            get { return _securityPair.Prices.OHLC.Close; }
            set { _securityPair.Prices.OHLC.Close = value; NotifyPropertyChanged("CloseB"); }
        }
        public double MidA
        {
            get { return _midA; }
            set { _midA = value; NotifyPropertyChanged("MidA"); }
        }
        public double MidB
        {
            get { return _midB; }
            set { _midB = value; NotifyPropertyChanged("MidB"); }
        }
        public double DayMoveA
        {
            get { return _moveDayA * 100; }
            set { _moveDayA = value; NotifyPropertyChanged("DayMoveA"); }
        }
        public double DayMoveB
        {
            get { return _moveDayB * 100; }
            set { _moveDayB = value; NotifyPropertyChanged("DayMoveB"); }
        }
        public double DayDifference
        {
            get { return _diffDay * 100; }
            private set { _diffDay = value; NotifyPropertyChanged("DayDifference"); }
        }
        public ushort Value
        {
            get { return _value; }
            set 
            {
                if(value <= UInt16.MaxValue) 
                    _value = value;
                NotifyPropertyChanged("Value"); 
            }
        }
        public double Trigger
        {
            get { return _trigger; }
            set
            {
                if (value > 0 && value <= 100)
                {
                    _trigger = value;
                    if (_trigger <= _exit)
                        Exit = 0;
                }
                else
                {
                    IsEnabled = false;
                    _trigger = 0;
                }
                NotifyPropertyChanged("Trigger");
            }
        }
        public double Exit
        {
            get { return _exit; }
            set
            {
                if (value >= 0 && value <= 100 && value < _trigger)
                {
                    _exit = value;
                }
                NotifyPropertyChanged("Exit");
            }
        }
    
        void  IMarketUpdateLast.OnPriceLast(IStock stock, Objects.Finance.Prices.PriceVolume last)
        {
            double mid;
            
            if (stock == _security)
            {
                mid = (_security.Prices.Ask.Price - _security.Prices.Bid.Price) + _security.Prices.Bid.Price;
                if (mid != _midA)
                {
                    MidA = mid;

                    DayMoveA = (mid - _security.Prices.OHLC.Close) / _security.Prices.OHLC.Close;

                    if (_midB != 0)
                    {
                        DayDifference = _moveDayA - _moveDayB;
                        CheckTriggerDay();
                    }                    
                }
            }
            else if (stock == _securityPair)
            {
                mid = (_securityPair.Prices.Ask.Price - _securityPair.Prices.Bid.Price) + _securityPair.Prices.Bid.Price;
                if (mid != _midB)
                {
                    MidB = mid;

                    DayMoveB = (mid - _securityPair.Prices.OHLC.Close) / _securityPair.Prices.OHLC.Close;

                    if (_midA != 0)
                    {
                        DayDifference = _moveDayA - _moveDayB;
                        CheckTriggerDay();
                    }              
                }
            }
        }



        private void NotifyPropertyChanged(string propertyName)
        {
            if (PropertyChanged != null)
            {
                PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
            }
        }


        private bool CheckTriggerDay()
        {
            if (!_isEnabled)
                return false;

            if (_trigger == 0 || _midA == 0 || _midB == 0)
                return false;

            if (_leg != null)
                return false;

            if (Math.Abs(_diffDay) * 100 > _trigger)
            {
                _leg = new FractionTrigger(this);
                Manager.TryAddTrigger(Guid.NewGuid(), _leg);
                NotifyPropertyChanged("IsTriggered");
                return true;
            }
            return false;
        }

        public void OnFractionTriggerComplete(FractionTrigger leg)
        {
            if (_leg == leg)
            {
                _leg = null;
                NotifyPropertyChanged("IsTriggered");
            }
        }

    }
}
